@@ -1,20 +1,39 @@
 """
 This is a vastly improved test runner over django's in-built runner.
 
-It provides the following advantages:
-  - Is not tied to application paths - you can choose to
-    run tests found in any module path.
-  - Prints exceptions as they occur, without waiting for
-    the entire test suite to complete as unittest does.
-  - assertEqual prints prettified output, so observing 
-    discrepancies should be easier.  This can be further
-    improved.
+Features:
+
+  - Exceptions and failures are reported instantly, and you're
+    given the option to launch a debugger at the point of failure.
+  - Pretty printing of stack traces -- frames involving your own
+    code are highlighted.
+  - Assert statements produce prettified output.
+  - Integrated selenium support:
+    - The test runner transparently starts its own self-contained (CherryPy)
+      server that the selenium tests access.
+    - We have a "--max-tries" argument that you can use to retry failed tests,
+      since selenium tests can be fragile and rely on timing.
+  - Support for separate workers if you want to run tests in parallel;
+    vastly improves test suite run times for selenium tests.
+  - Can run tests by partial paths rather than full paths (eg. can specify
+    just the test _method_ name from the command line, rather than the
+    full module path).
+  - Test tagging.
+  - Coverage integration.
+  - Produces XML reports of 
+  - Some hudson integration: can choose to run all the tests that failed 
+    in a particular build.
+  - Support for "variations" on tests: a way to slightly change tests
+    with a minimum of code.
 
-We also have integrated selenium support.
 
 Usage:
-  - Make your test classes derive from django_test.TestCase
-  - Set your settings.TEST_RUNNER to django_test.run_tests
+  - Add 'django_test' to your settings.INSTALLED_APPS
+    - This will replace the `manage.py test` command with `django_test`'s,
+      which is mostly compatible with django's but has many improvements.
+
+  - Make your test classes derive from django_test.TestCase (or django_test.selenium.SeleniumTestCase)
+
   - Run python manage.py test as normal:
 
     eg:
@@ -25,7 +44,7 @@ Usage:
   - Settings in settings.TEST_SETTINGS dictionary will override normal settings
   - To run Selenium tests, set RUN_SELENIUM_TESTS to True
 
-Limitations:
+TODO:
   - Currently, only classes ending with "Test" and methods
     starting with "test_" are found when looking for tests.
   - We don't have the full set of assert* functions.
@@ -47,13 +66,11 @@ from django.conf import settings
 from django.db import connection, transaction
 from django.test.testcases import disable_transaction_methods, restore_transaction_methods
 from django.core.urlresolvers import clear_url_caches
-from django_test.xmlreport import XmlTestSuite
 from django.contrib.sites.models import Site
 from django.db.models.query import QuerySet
 from django.db import connections, DEFAULT_DB_ALIAS
 from django.core.management import call_command
 from django.test.utils import setup_test_environment, teardown_test_environment
-from optparse import OptionParser
 
 from django_test.utils import import_object, import_module, restore_database
 from django_test.variations import MyVariationRunner
@@ -64,18 +81,78 @@ class OptionDescriptor(object):
     self.args = args
     self.kwargs = kwargs
 
-COMMAND_LINE_OPTIONS = (
+
+PLUGINS = [import_object(path)() for path in settings.TEST_PLUGINS]
+
+command_line_options = [
   OptionDescriptor("-g", "--grep", dest = "grep", action = "append"),
-  OptionDescriptor("-f", "--failed", dest = "failed", action = "store"),
   OptionDescriptor("-t", "--tag", dest = "tag", action = "store"),
-  OptionDescriptor("-m", "--max_tries", dest = "max_tries", action = "store", 
+  OptionDescriptor("-m", "--max-tries", dest = "max_tries", action = "store", 
                          type = "int", default = 1),
+  OptionDescriptor("-n", "--never-debug", dest = "never_debug", action = "store_true", 
+                         default = False),
   #TODO:
   #make_option('--numworkers', '-n', dest='num_workers',
   #    help='The number of workers.'),
   #make_option('--workernum', '-w', dest='worker_num',
   #    help='The number of this worker [0..numworkers-1].'),
-)
+]
+
+for p in PLUGINS:
+  command_line_options.extend(p.get_command_line_options())
+  
+
+def for_all_plugins(method_name, *args, **kwargs):
+  for p in PLUGINS:
+    getattr(p, method_name)(*args, **kwargs)
+
+
+
+class Filter(object):
+  def test_class(self, klass):
+    return True
+
+  def test_method(self, klass, method):
+    return True
+
+
+class GrepFilter(Filter):
+  def __init__(self, string):
+    self.string = string
+
+  def test_class(self, klass):
+    return any((self.string in get_method_path(klass, n) \
+                for n in dir(klass) \
+                if n.startswith("test_")))
+
+  def test_method(self, klass, method):
+    return self.string in get_method_path(klass, method)
+
+
+class TagFilter(Filter):
+  def __init__(self, tag):
+    self.tag = tag
+
+  def test_class(self, klass):
+    return any((self.tag in getattr(getattr(klass, n, None), 'tags', [])
+                for n in dir(klass) \
+                if n.startswith("test_")))
+
+  def test_method(self, klass, method):
+    return self.tag in getattr(method, 'tags', [])
+
+
+class NoSeleniumFilter(Filter):
+  def test_class(self, klass):
+    return not getattr(klass, 'selenium_test', False)
+
+
+def get_method_path(klass, method):
+  if hasattr(method, "__name__"):
+    method = method.__name__
+
+  return "%s.%s.%s" % (klass.__module__, klass.__name__, method)
+
 
 
 class TestDriver(object):
@@ -83,7 +160,9 @@ class TestDriver(object):
                test_paths = None, 
                num_workers = None, 
                worker_num = None,
-               max_tries = 1):
+               max_tries = 1,
+               never_debug = False,
+               run_selenium_tests = True):
     self.test_paths = test_paths
     self.num_workers = num_workers \
                        if num_workers != None \
@@ -93,18 +172,22 @@ class TestDriver(object):
                       else getattr(settings, "WORKER_NUM", 0)
     self.verbosity = 1
     self.max_tries = max_tries
+    self.never_debug = never_debug
 
     self.passed = 0
     self.errors = 0
     self.tests_seen = 0
     self.tests_run = 0
 
-    #TODO: factor this
-    self.suite = XmlTestSuite(self.worker_num)    
+    for_all_plugins("init", self)
+
     self.modules = []
     self.classes = []
     self.methods = []
-    self.filters = []
+
+    self.filters = []  # List of Filter objects
+    if not run_selenium_tests:
+      self.filters.append(NoSeleniumFilter())
 
 
   def add_path(self, path):
@@ -135,10 +218,6 @@ class TestDriver(object):
 
 
   def run(self):
-    #TODO: factor this
-    from playfi.util.debug import listen_for_sigquit
-    listen_for_sigquit()
-
     setup_test_environment()
     settings.DEBUG = False
     old_name = settings.DATABASES["default"]["NAME"]
@@ -171,9 +250,6 @@ class TestDriver(object):
   def add_method(self, klass, method_name):
     self.methods.append((klass, method_name))
 
-  def add_filter(self, predicate):
-    self.filters.append(predicate)
-
   def test_method(self, class_or_object, method_name):
     if inspect.isclass(class_or_object):
       klass = class_or_object
@@ -183,57 +259,57 @@ class TestDriver(object):
       ob = class_or_object
 
     method = ob.__getattribute__(method_name)
-    method_path = "%s:%s.%s" % (klass.__module__, klass.__name__, method_name)
+    method_path = "%s.%s.%s" % (klass.__module__, klass.__name__, method_name)
 
     self.tests_seen += 1
 
     if self.tests_seen % self.num_workers != self.worker_num:
       return
 
-    if self.filters and not any([p(klass, method, method_path) for p in self.filters]):
+    if self.filters and any((not p.test_method(klass, method) for p in self.filters)):
       return
 
     if getattr(method, "long_test", False) and not getattr(settings, "RUN_LONG_TESTS", "True"):
       print "Skipping test %s as it's long and we're not running long tests."
       return
 
-    result = self.suite.test(klass.__module__, klass.__name__, method_name)
 
-    print >>sys.stderr, "-" * 50
-    print >>sys.stderr, "%s (Worker %s/%s)" % \
-                        (method_path, self.worker_num, self.num_workers)
-    print >>sys.stderr, "-" * 50
+    print "-" * 50
+    print "%s (Worker %s/%s)" % \
+          (method_path, self.worker_num, self.num_workers)
+    print "-" * 50
 
     try_num = 1
     finished = False
     torn_down = False
-    result.start()
+
+    for_all_plugins("start_running_test", self, klass, method_name)
 
     while not finished and try_num <= self.max_tries:
       try:
         # Actually run the test
         if getattr(method, "has_variations", False):
-          self.coverage.start()
+          for_all_plugins("start_running_test_try", self, klass, method_name)
           MyVariationRunner().run_test(method)
-          self.coverage.stop()
+          for_all_plugins("finish_running_test_try", self, klass, method_name)
           torn_down = True
         else:
           ob._test_setup()
           ob.setUp()
-          self.coverage.start()
+          for_all_plugins("start_running_test_try", self, klass, method_name)
           method()
-          self.coverage.stop()
+          for_all_plugins("finish_running_test_try", self, klass, method_name)
 
-        if hasattr(ob, "post_test_checks"):
-          ob.post_test_checks(method)
-
-        result.success()
+        for_all_plugins("post_test_checks", self, klass, method)
+        for_all_plugins("test_success", self, klass, method)
         finished = True
 
         print "PASSED"
         self.passed += 1
         self.tests_run += 1
       except Exception, e:
+        for_all_plugins("finish_running_test_try", self, klass, method_name)
+
         if try_num == self.max_tries:
           print "FAILED: ", e
           etype, value, tb = sys.exc_info()
@@ -241,15 +317,14 @@ class TestDriver(object):
           st = traceback.format_tb(tb)
           ex = traceback.format_exception_only(etype, value)
           
-          result.failure(''.join(ex), ''.join(st))      
+          for_all_plugins("test_failure", self, klass, method_name, ex, st)
           
           for line in itertools.chain(st, ex):
             print re.sub('(' + os.getcwd() + ')(.*)", (line \\d+)', 
                          '\\1\033[1;31m\\2\033[0;0;0m", \033[1;32m\\3\033[0;0;0m', 
                          line)
 
-          if not (getattr(self, 'never_debug', False) or 
-                  getattr(settings, "NEVER_DEBUG", False)):
+          if not self.never_debug:
             print "Would you like to debug? [y/N/neVer]"
             c = raw_input()
             if c.lower() == "y":
@@ -290,20 +365,14 @@ class TestDriver(object):
     for module_object_name in dir(module):
       module_object = getattr(module, module_object_name)
       if inspect.isclass(module_object):
-        if issubclass(module_object, TestCase):
-          #TODO: factor this
-          from django_test.selenium import SeleniumTestCase
-          if not issubclass(module_object, SeleniumTestCase) or self.run_selenium_tests:
-            self.test_class(module_object)
+        klass = module_object
+        if issubclass(klass, TestCase):
+          if self.filters and any((not f.test_class(klass) for f in self.filters)):
+            continue
+          self.test_class(klass)
 
 
   def run_tests(self):
-    #TODO: factor this
-    from coverage import coverage
-    self.coverage = coverage()
-
-    self.run_selenium_tests = getattr(settings, "RUN_SELENIUM_TESTS", False)
-    
     for m in self.modules:
       self.test_module(m)
 
@@ -316,36 +385,34 @@ class TestDriver(object):
       if getattr(instance, "post_last_teardown", None):
         instance.post_last_teardown()
 
-    #TODO: factor this
-    if hasattr(settings, "XML_REPORT_DIR"):
-      self.suite.generate_xml(settings.XML_REPORT_DIR)
-
-    #TODO: factor this
-    self.coverage.html_report(directory = 'covhtml')
-
+    for_all_plugins("finish_running_tests", self)
 
 
 class CustomTestRunner(object):
   def run_tests(self, *test_paths, **options):
     driver = TestDriver(max_tries = options['max_tries'])
 
+    driver.never_debug = options["never_debug"]
+
     if test_paths:
       for p in test_paths:
         driver.add_path(p)
     elif options['grep']:
       driver.add_all_tests()
-      for p in options['grep']:
-        driver.add_filter(lambda _1, _2, method_path: p in method_path)
-    elif options['failed']:
-      #TODO: factor this
-      from playfi.util.hudson import get_failed_test_names
-      for p in get_failed_test_names(int(['options'].failed)):
-        driver.add_path(p)
+      for g in options['grep']:
+        driver.filters.append(GrepFilter(g))
     elif options['tag']:
       driver.add_all_tests()
-      driver.add_filter(lambda _1, method, _: options['tag'] in getattr(method, 'tags', []))
+      driver.filters.append(GrepFilter(options['tag']))
     else:
-      driver.add_all_tests()
+      add_all_tests = True
+      for p in PLUGINS:
+        handled = p.process_command_line_options(driver, **options)
+        if handled:
+          add_all_tests = False
+          break
+      if add_all_tests:
+        driver.add_all_tests()
       
     driver.run()
 
@@ -465,16 +532,6 @@ class TestCase(object):
     self._teardown_urlconf()
 
 
-  def post_test_checks(self, method):
-    if not getattr(method, "allow_server_errors", False):
-      #TODO: factor this
-      from playfi.exception_middleware.models import ErrorReport
-      # Do an actual list comparison so that if it fails, assertEqual
-      # will print out the exception messages.
-      self.assertEqual(list(ErrorReport.objects.all()), [])
-
-
-
 
 
 
